@@ -24,7 +24,8 @@ from agora.model.transaction import TransactionData
 from agora.model.transaction_type import TransactionType
 from agora.retry import retry, LimitStrategy, BackoffWithJitterStrategy, BinaryExponentialBackoff, \
     NonRetriableErrorsStrategy, RetriableErrorsStrategy
-from agora.utils import partition, quarks_to_kin
+from agora.utils import partition, quarks_to_kin, user_agent
+from agora.version import VERSION
 
 _SUPPORTED_VERSIONS = [3]
 
@@ -129,14 +130,14 @@ class BaseClient:
         raise NotImplementedError("BaseClient is an abstract class. Subclasses must implement submit_payment")
 
     def submit_earn_batch(
-        self, sender: PrivateKey, earns: List[Earn], source: Optional[PrivateKey] = None, memo: Optional[str] = None
+        self, sender: PrivateKey, earns: List[Earn], channel: Optional[PrivateKey] = None, memo: Optional[str] = None
     ) -> BatchEarnResult:
         """Submit multiple earn payments.
 
         :param sender: The :class:`PrivateKey <agora.model.keys.PrivateKey` of the sender
         :param earns: A list of :class:`Earn <agora.model.earn.Earn>` objects.
-        :param source: (optional) The :class:`PrivateKey <agora.model.keys.PrivateKey` of the transaction source
-            account. If not set, the sender will be used as the source.
+        :param channel: (optional) The :class:`PrivateKey <agora.model.keys.PrivateKey` of a channel account to use as
+            the transaction source. If not set, the `sender` will be used as the source.
         :param memo: (optional) The memo to include in the transaction. If set, none of the invoices included in earns
             will be applied.
 
@@ -146,6 +147,12 @@ class BaseClient:
         """
         raise NotImplementedError("BaseClient is an abstract class. Subclasses must implement submit_earn_batch")
 
+    def close(self) -> None:
+        """Closes the connection-related resources (e.g. the gRPC channel) used by the client. Subsequent requests to
+        this client will cause an exception to be thrown.
+        """
+        raise NotImplementedError("BaseClient is an abstract class. Subclasses must implement close")
+
 
 class SubmitStellarTransactionResult:
     def __init__(self, tx_hash: Optional[bytes] = None,
@@ -186,10 +193,12 @@ class Client(BaseClient):
         if not grpc_channel:
             endpoint = endpoint if endpoint else _ENDPOINTS[env]
             ssl_credentials = grpc.ssl_channel_credentials()
-            grpc_channel = grpc.secure_channel(endpoint, ssl_credentials)
+            self.grpc_channel = grpc.secure_channel(endpoint, ssl_credentials)
+        else:
+            self.grpc_channel = grpc_channel
 
-        self.account_stub = account_pb_grpc.AccountStub(grpc_channel)
-        self.transaction_stub = tx_pb_grpc.TransactionStub(grpc_channel)
+        self.account_stub = account_pb_grpc.AccountStub(self.grpc_channel)
+        self.transaction_stub = tx_pb_grpc.TransactionStub(self.grpc_channel)
 
         self.whitelist_key = whitelist_key
 
@@ -211,6 +220,8 @@ class Client(BaseClient):
         # but it does not get used to submit transactions
         self._horizon = kin_base.Horizon()
 
+        self._metadata = (user_agent(VERSION),)
+
         # Since we don't actually use Horizon for any requests, call `self._horizon.close()` to preemptively ensure that
         # any open aiohttp.ClientSessions get closed.
         loop = asyncio.new_event_loop()
@@ -228,7 +239,7 @@ class Client(BaseClient):
             transaction_hash=model_pb2.TransactionHash(
                 value=tx_hash
             )
-        ), timeout=_GRPC_TIMEOUT_SECONDS)
+        ), metadata=self._metadata, timeout=_GRPC_TIMEOUT_SECONDS)
 
         if resp.state is tx_pb.GetTransactionResponse.State.UNKNOWN:
             raise TransactionNotFound()
@@ -251,7 +262,7 @@ class Client(BaseClient):
         if payment.invoice and self.app_index <= 0:
             raise ValueError("cannot submit a payment with an invoice without an app index")
 
-        builder = self._get_stellar_builder(payment.source if payment.source else payment.sender)
+        builder = self._get_stellar_builder(payment.channel if payment.channel else payment.sender)
 
         invoice_list = None
         if payment.memo:
@@ -261,7 +272,7 @@ class Client(BaseClient):
                 invoice_list = InvoiceList(invoices=[payment.invoice])
 
             fk = invoice_list.get_sha_224_hash() if payment.invoice else b''
-            memo = AgoraMemo.new(1, payment.payment_type, self.app_index, fk)
+            memo = AgoraMemo.new(1, payment.tx_type, self.app_index, fk)
             builder.add_hash_memo(memo.val)
 
         builder.append_payment_op(
@@ -270,8 +281,8 @@ class Client(BaseClient):
             source=payment.sender.public_key.stellar_address,
         )
 
-        if payment.source:
-            signers = [payment.source, payment.sender]
+        if payment.channel:
+            signers = [payment.channel, payment.sender]
         else:
             signers = [payment.sender]
 
@@ -305,7 +316,7 @@ class Client(BaseClient):
         return result.tx_hash
 
     def submit_earn_batch(
-        self, sender: PrivateKey, earns: List[Earn], source: Optional[bytes] = None, memo: Optional[str] = None
+        self, sender: PrivateKey, earns: List[Earn], channel: Optional[bytes] = None, memo: Optional[str] = None
     ) -> BatchEarnResult:
         if self._kin_version not in _SUPPORTED_VERSIONS:
             raise UnsupportedVersionError
@@ -323,7 +334,7 @@ class Client(BaseClient):
         failed = []
         for earn_batch in partition(earns, 100):
             try:
-                result = self._submit_earn_batch_tx(sender, earn_batch, source, memo)
+                result = self._submit_earn_batch_tx(sender, earn_batch, channel, memo)
             except Error as e:
                 failed += [EarnResult(earn, error=e) for idx, earn in enumerate(earn_batch)]
                 break
@@ -348,16 +359,19 @@ class Client(BaseClient):
 
         return BatchEarnResult(succeeded=succeeded, failed=failed)
 
+    def close(self) -> None:
+        self.grpc_channel.close()
+
     def _submit_earn_batch_tx(
-        self, sender: PrivateKey, earns: List[Earn], source: Optional[PrivateKey] = None, memo: Optional[str] = None
+        self, sender: PrivateKey, earns: List[Earn], channel: Optional[PrivateKey] = None, memo: Optional[str] = None
     ) -> SubmitStellarTransactionResult:
         """ Submits a single transaction for a batch of earns. An error will be raised if the number of earns exceeds
         the capacity of a single transaction.
 
         :param sender: The :class:`PrivateKey <agora.model.keys.PrivateKey` of the sender
         :param earns: A list of :class:`Earn <agora.model.earn.Earn>` objects.
-        :param source: (optional) The :class:`PrivateKey <agora.model.keys.PrivateKey` of the transaction source
-            account. If not set, the sender will be used as the source.
+        :param channel: (optional) The :class:`PrivateKey <agora.model.keys.PrivateKey` of the channel account to use
+            as the transaction source. If not set, the sender will be used as the source.
         :param memo: (optional) The memo to include in the transaction. If set, none of the invoices included in earns
             will be applied.
 
@@ -366,7 +380,7 @@ class Client(BaseClient):
         if len(earns) > 100:
             raise ValueError("cannot send more than 100 earns")
 
-        builder = self._get_stellar_builder(source if source else sender)
+        builder = self._get_stellar_builder(channel if channel else sender)
 
         invoices = [earn.invoice for earn in earns if earn.invoice]
         invoice_list = InvoiceList(invoices) if invoices else None
@@ -384,8 +398,8 @@ class Client(BaseClient):
                 source=sender.public_key.stellar_address,
             )
 
-        if source:
-            signers = [source, sender]
+        if channel:
+            signers = [channel, sender]
         else:
             signers = [sender]
 
@@ -402,16 +416,23 @@ class Client(BaseClient):
     def _sign_and_submit_builder(
         self, signers: List[PrivateKey], builder: kin_base.Builder, invoice_list: Optional[InvoiceList] = None
     ) -> SubmitStellarTransactionResult:
-        def _sign_and_submit():
-            builder.te = None  # reset the envelope
-            source_info = self._get_stellar_account_info(signers[0].public_key)
-            builder.sequence = source_info.sequence_number + 1
+        source_info = self._get_stellar_account_info(signers[0].public_key)
+        offset = 1
 
+        def _sign_and_submit():
+            nonlocal offset
+
+            # reset generated tx and te
+            builder.tx = None
+            builder.te = None
+
+            builder.sequence = source_info.sequence_number + offset
             for signer in signers:
                 builder.sign(signer.stellar_seed)
 
             result = self._submit_stellar_transaction(base64.b64decode(builder.gen_xdr()), invoice_list)
             if result.tx_error and isinstance(result.tx_error.tx_error, BadNonceError):
+                offset += 1
                 raise result.tx_error.tx_error
 
             return result
@@ -426,8 +447,8 @@ class Client(BaseClient):
         resp = self.account_stub.CreateAccount(account_pb.CreateAccountRequest(
             account_id=model_pb2.StellarAccountId(
                 value=private_key.public_key.stellar_address
-            )
-        ), timeout=_GRPC_TIMEOUT_SECONDS)
+            ),
+        ), metadata=self._metadata, timeout=_GRPC_TIMEOUT_SECONDS)
         if resp.result == account_pb.CreateAccountResponse.Result.EXISTS:
             raise AccountExistsError()
 
@@ -440,8 +461,8 @@ class Client(BaseClient):
         resp = self.account_stub.GetAccountInfo(account_pb.GetAccountInfoRequest(
             account_id=model_pb2.StellarAccountId(
                 value=public_key.stellar_address
-            )
-        ), timeout=_GRPC_TIMEOUT_SECONDS)
+            ),
+        ), metadata=self._metadata, timeout=_GRPC_TIMEOUT_SECONDS)
         if resp.result == account_pb.GetAccountInfoResponse.Result.NOT_FOUND:
             raise AccountNotFoundError
 
@@ -477,7 +498,7 @@ class Client(BaseClient):
                 envelope_xdr=tx_bytes,
                 invoice_list=invoice_list.to_proto() if invoice_list else None,
             )
-            resp = self.transaction_stub.SubmitTransaction(req, timeout=_GRPC_TIMEOUT_SECONDS)
+            resp = self.transaction_stub.SubmitTransaction(req, metadata=self._metadata, timeout=_GRPC_TIMEOUT_SECONDS)
 
             result = SubmitStellarTransactionResult(tx_hash=resp.hash.value)
             if resp.result == tx_pb.SubmitTransactionResponse.Result.REJECTED:
